package com.flyfish.framework.beans.meta;

import com.fasterxml.jackson.annotation.JsonFormat;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.flyfish.framework.annotations.Properties;
import com.flyfish.framework.annotations.*;
import com.flyfish.framework.beans.enums.ValidationCandidate;
import com.flyfish.framework.domain.base.Qo;
import com.flyfish.framework.domain.base.Vo;
import com.flyfish.framework.utils.ReflectionUtils;
import com.flyfish.framework.utils.StringFormats;
import com.flyfish.framework.validation.annotations.Money;
import lombok.Data;
import lombok.val;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.ListUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.ClassUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.reflect.FieldUtils;
import org.apache.commons.lang3.reflect.TypeUtils;
import org.springframework.beans.BeanUtils;
import org.springframework.core.annotation.AnnotatedElementUtils;
import org.springframework.core.annotation.MergedAnnotation;
import org.springframework.core.annotation.MergedAnnotations;
import org.springframework.data.mongodb.core.mapping.Document;

import java.beans.PropertyDescriptor;
import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.*;
import java.util.stream.Collectors;

/**
 * bean的属性描述
 * 对应js的类型
 *
 * @author wangyu
 */
@Data
public class BeanProperty {

    // 属性名称
    private String name;

    // 属性标题
    private String title;

    // 旧标题
    @JsonIgnore
    private transient String oldTitle;

    // 描述信息
    private String description;

    // 布局
    private String layout;

    // bean属性的类型，js类型
    private BeanPropertyType type;

    // 只读属性，不生成于表单
    @JsonIgnore
    private transient boolean readonly;

    // 继承属性，不生成于表单
    @JsonIgnore
    private transient boolean inherited;

    // 属性
    private Map<String, Object> props;

    // 视图选项
    private Map<String, Object> options;

    // 类型为object时，拥有子表单
    private List<BeanProperty> children;

    // bean验证
    private List<BeanValidation> validation;

    // 排序变量，用于排序
    @JsonIgnore
    private transient int order;

    // 所属分组，该分组必须在元数据中定义
    private String group;

    // 其他属性，可交由前端展开
    private Map<String, Object> extra = new HashMap<>();

    /**
     * 来自属性解释器构造
     * 支持对象嵌套
     *
     * @param descriptor 解释器
     * @return 结果
     */
    public static BeanProperty form(PropertyDescriptor descriptor, Class<?> beanClass, Object instance) {
        // 组装属性
        BeanProperty property = new BeanProperty();
        property.setName(descriptor.getName());
        property.setType(BeanPropertyType.of(descriptor, beanClass));
        if (null == property.getType()) {
            property.setReadonly(true);
            return property;
        }
        // 如果是qo或者vo，处于严格模式
        boolean strict = Qo.class.isAssignableFrom(beanClass) || Vo.class.isAssignableFrom(beanClass);
        // 尝试获取field
        Field field = FieldUtils.getField(beanClass, descriptor.getName(), true);
        MergedAnnotations annotations = null == field ? MergedAnnotations.from() : MergedAnnotations.from(field);
        // 存在field，可以干一些坏事
        if (null != field) {
            // 开始解析关键注解
            if (annotations.isPresent(Property.class)) {
                Property props = annotations.get(Property.class).synthesize();
                // 只有存在关键注解，才会进行初始化
                String parentName = Optional.ofNullable(beanClass.getAnnotation(RestBean.class))
                        .map(RestBean::name).orElse("");
                property.setDescription(props.description());
                property.setReadonly(props.readonly());
                property.setInherited(props.inherited());
                property.setGroup(props.group());
                // 继承模式，继承名称
                if (property.inherited) {
                    property.setOldTitle(props.title());
                    property.setTitle(parentName + property.oldTitle);
                } else {
                    property.setTitle(props.title());
                }
                // 优雅地设置排序
                MergedAnnotation<Order> order = annotations.get(Order.class);
                if (order.isPresent()) {
                    property.setOrder(order.synthesize().value());
                } else {
                    property.setOrder(props.order());
                }
                // 追加生成属性
                MergedAnnotation<Generation> generation = annotations.get(Generation.class);
                if (generation.isPresent()) {
                    property.extra.put(BeanProps.GENERATED, generation.asMap());
                    property.extra.put(BeanProps.COMPONENT, "input-hidden");
                }
                // 追加属性映射
                MergedAnnotation<ComputedProps.List> links = annotations.get(ComputedProps.List.class);
                if (links.isPresent()) {
                    property.extra.put(BeanProps.LINKED, Arrays.stream(
                            links.getAnnotationArray("value", ComputedProps.class)
                    ).map(MergedAnnotation::asMap).collect(Collectors.toList()));
                } else {
                    MergedAnnotation<ComputedProps> single = annotations.get(ComputedProps.class);
                    if (single.isPresent()) {
                        property.extra.put(BeanProps.LINKED, Collections.singletonList(links.asMap()));
                    }
                }
                // 优雅的设置额外的属性
                MergedAnnotation<FormItem> item = annotations.get(FormItem.class);
                if (item.isPresent()) {
                    applyFormItem(property, item.synthesize());
                }
                // 优雅的设置联动映射
                MergedAnnotation<MappedTo> mapping = annotations.get(MappedTo.class);
                if (mapping.isPresent()) {
                    property.extra.put(BeanProps.MAPPING, mapping.asMap());
                }
                // 处理联动
                MergedAnnotation<ConditionOn.List> condition = annotations.get(ConditionOn.List.class);
                if (condition.isPresent()) {
                    property.extra.put(BeanProps.CONDITION, Arrays.stream(
                            condition.getAnnotationArray("value", ConditionOn.class)
                    ).map(MergedAnnotation::asMap).collect(Collectors.toList()));
                } else {
                    MergedAnnotation<ConditionOn> single = annotations.get(ConditionOn.class);
                    if (single.isPresent()) {
                        property.extra.put(BeanProps.CONDITION, Collections.singletonList(single.asMap()));
                    }
                }
                // 优雅的处理校验
                property.setValidation(ValidationCandidate.produce(annotations, property.getType()));
            } else if (strict) {
                property.setReadonly(true);
                return property;
            }
        } else if (strict) {
            property.setReadonly(true);
            return property;
        }
        Class<?> clazz = descriptor.getPropertyType();
        switch (property.getType()) {
            case STRING:
                // 如果字段使用DictValue注解，尝试使用字典值仓库
                if (null != field) {
                    // 添加了字典枚举，自动添加code，从字典表取得
                    if (annotations.isPresent(DictValue.class)) {
                        DictValue dictValue = annotations.get(DictValue.class).synthesize();
                        property.prop("code", dictValue.value());
                    } else if (annotations.isPresent(EnumValue.class)) {
                        // 添加了枚举注解，自动注入类型，给前端使用
                        property.setType(BeanPropertyType.ENUM);
                        EnumValue enumValue = annotations.get(EnumValue.class).synthesize();
                        String name = StringFormats.camel2Line(ClassUtils.getShortClassName(enumValue.value()));
                        property.prop("code", name);
                    } else if (annotations.isPresent(DBRefValue.class)) {
                        // 添加了数据库引用注解，自动注入类型给前端使用
                        DBRefValue dbRefValue = annotations.get(DBRefValue.class).synthesize();
                        Optional<String> optional = processDbRef(dbRefValue.value());
                        if (optional.isPresent()) {
                            property.setType(BeanPropertyType.DB_REF);
                            property.prop("uri", optional.get());
                        } else {
                            property.setType(BeanPropertyType.STRING);
                        }
                    }
                }
                break;
            case NUMBER:
                if (null != field) {
                    if (annotations.isPresent(Money.class) && !property.extra.containsKey(BeanProps.COMPONENT)) {
                        property.extra.put(BeanProps.COMPONENT, "money-input");
                    }
                }
                break;
            case LIST:
                if (null != field) {
                    // 是附件，展现附件列表
                    ReflectionUtils.getGenericType(field)
                            .filter(property::isAttachment)
                            .ifPresent(item -> property.prop("attachment", true));
                    if (annotations.isPresent(SubBean.class)) {
                        // 尝试获取泛型参数，存在时，赋值子表单
                        parseSubClass(field).ifPresent(subClazz -> property.setChildren(from(subClazz)));
                        // 处理替换
                        applyReplacement(property, annotations);
                    } else if (annotations.isPresent(DateRange.class)) {
                        property.setType(BeanPropertyType.DATE);
                        property
                                .prop("type", "range")
                                .prop("placeholder", Arrays.asList("开始时间", "结束时间"));
                    } else if (annotations.isPresent(DBRefValue.class)) {
                        // 添加了数据库引用注解，自动注入类型给前端使用
                        DBRefValue dbRefValue = annotations.get(DBRefValue.class).synthesize();
                        processDbRef(dbRefValue.value()).ifPresent(ref -> {
                            property.setType(BeanPropertyType.DB_REF);
                            property.prop("uri", ref).prop("mode", "multiple");
                        });
                    }
                }
                break;
            case OBJECT:
                // 有子bean注解才处理
                if (null != field && annotations.isPresent(SubBean.class)) {
                    property.setChildren(from(clazz));
                    // 处理替换
                    applyReplacement(property, annotations);
                }
                break;
            case DB_REF:
                // 当存在db-ref时，解析为动态数据源
                Optional<String> optional = processDbRef(clazz);
                if (optional.isPresent()) {
                    if (property.isAttachment(descriptor.getPropertyType())) {
                        property.prop("attachment", true);
                    } else {
                        property.prop("uri", optional.get());
                    }
                } else {
                    property.setType(BeanPropertyType.STRING);
                }
                break;
            case ENUM:
                // 为枚举时，填入枚举类型
                String name = StringFormats.camel2Line(ClassUtils.getShortClassName(clazz));
                property.prop("code", name);
                break;
            case DATE:
                // 为日期，自动放入类型和长度
                if (null != field) {
                    if (annotations.isPresent(JsonFormat.class)) {
                        String pattern = annotations.get(JsonFormat.class).synthesize().pattern();
                        // yyyy-MM-dd HH:mm:ss
                        if (StringUtils.isNotBlank(pattern)) {
                            int length = pattern.length();
                            if (length == 7) {
                                property.prop("type", "month");
                            } else if (length == 10) {
                                property.prop("type", "date");
                            } else if (length == 11) {
                                property.prop("type", "date")
                                        .prop("format", "YYYY年MM月DD日");
                            } else if (length == 16) {
                                property.prop("type", "datetime");
                                property.prop("format", "YYYY-MM-DD HH:mm");
                            } else if (length > 16) {
                                property.prop("format", "YYYY-MM-DD HH:mm:ss");
                            }
                        }
                    }
                }
        }
        // 写入默认值
        Object value = ReflectionUtils.getFieldValue(instance, property.name);
        if (null != value) {
            property.option("initialValue", value);
        }
        return property;
    }

    /**
     * 类加载并解析属性
     *
     * @return 结果
     */
    public static List<BeanProperty> from(Class<?> clazz) {
        Object instance = ReflectionUtils.instantiate(clazz);
        PropertyDescriptor[] descriptors = BeanUtils.getPropertyDescriptors(clazz);
        List<BeanProperty> properties = Arrays.stream(descriptors)
                .filter(descriptor -> !"class".equals(descriptor.getName()))
                .map(descriptor -> BeanProperty.form(descriptor, clazz, instance))
                .filter(property -> !property.isReadonly())
                .collect(Collectors.toList());
        // 这里进行一些修复和初始化
        List<BeanProperty> result = ListUtils.union(parseExtras(clazz, properties), properties);
        // 排序
        result.sort(Comparator.comparingInt(a -> a.order));
        return result;
    }

    /**
     * 解析db引用的参数
     *
     * @param clazz 类
     * @return 结果
     */
    private static Optional<String> processDbRef(Class<?> clazz) {
        return Optional.ofNullable(AnnotatedElementUtils.findMergedAnnotation(clazz, Document.class))
                .map(Document::value);
    }

    /**
     * 解析额外的注解
     * 以该注解的优先级更高
     *
     * @param clazz  类
     * @param origin 原本的元数据
     * @return 结果
     */
    private static List<BeanProperty> parseExtras(Class<?> clazz, List<BeanProperty> origin) {
        List<BeanProperty> result = new ArrayList<>();
        // 包含properties
        if (clazz.isAnnotationPresent(Properties.class)) {
            Properties properties = clazz.getAnnotation(Properties.class);
            for (Property prop : properties.value()) {
                applyExtraProperty(result, origin, prop);
            }
        } else if (clazz.isAnnotationPresent(Property.class)) {
            applyExtraProperty(result, origin, clazz.getAnnotation(Property.class));
        }
        return result;
    }

    /**
     * 解析子类型，用于表格组件
     *
     * @param field 反射域
     * @return 结果
     */
    private static Optional<Class<?>> parseSubClass(Field field) {
        // 获取实际的泛型
        Type type = field.getGenericType();
        if (type instanceof ParameterizedType) {
            val types = TypeUtils.getTypeArguments((ParameterizedType) type);
            return types.entrySet().stream().findFirst().map(entry -> {
                Type parsed = entry.getValue();
                if (parsed instanceof Class) {
                    Class<?> clazz = (Class<?>) parsed;
                    if (!ClassUtils.isPrimitiveOrWrapper(clazz) && !CharSequence.class.isAssignableFrom(clazz)) {
                        return clazz;
                    }
                }
                return null;
            });
        }
        return Optional.empty();
    }

    /**
     * 将formItem注解赋值到目标属性上
     *
     * @param property 目标属性
     * @param item     注解元
     */
    private static void applyFormItem(BeanProperty property, FormItem item) {
        // 设置额外属性
        if (StringUtils.isNotBlank(item.component())) {
            property.extra.put(BeanProps.COMPONENT, item.component());
        }
        if (StringUtils.isNotBlank(item.help())) {
            property.extra.put(BeanProps.HELP, item.help());
        }
        if (StringUtils.isNotBlank(item.layout())) {
            property.layout = item.layout();
        }
        if (ArrayUtils.isNotEmpty(item.props())) {
            for (FormItem.Prop prop : item.props()) {
                property.prop(prop.key(), prop.value());
            }
        }
    }

    /**
     * 使的额外的属性生效
     *
     * @param result 结果集
     * @param origin 原属性集
     * @param prop   当前注解
     */
    private static void applyExtraProperty(List<BeanProperty> result, List<BeanProperty> origin, Property prop) {
        String key = prop.key();
        // 如果基础元数据已经包含了这部分，相当于重载属性，此时不会调整其他属性，而是替换名称（将来可能扩展更多）
        Optional<BeanProperty> found = origin.stream().filter(property -> property.name.equals(key))
                .findFirst();
        if (found.isPresent()) {
            BeanProperty property = found.get();
            if (prop.order() != 0) {
                property.order = prop.order();
            }
            if (StringUtils.isNotBlank(prop.group())) {
                property.group = prop.group();
            }
            // 只读状态不显示
            if (prop.readonly()) {
                origin.remove(property);
            }
            // 存在校验，替换名称文案(仅inherit)
            if (property.inherited && CollectionUtils.isNotEmpty(property.validation)) {
                property.validation.forEach(validation -> validation.setMessage(validation.getMessage()
                        .replace(property.oldTitle, prop.title())));
            }
            // 处理额外的表单属性
            if (ArrayUtils.isNotEmpty(prop.form())) {
                applyFormItem(property, prop.form()[0]);
            }
            // 处理额外的映射
            if (ArrayUtils.isNotEmpty(prop.mapping())) {
                property.extra.put(BeanProps.MAPPING, MergedAnnotations.from(prop.mapping()).get(MappedTo.class).asMap());
            }
            // 处理值生成策略，此处可清除原策略
            if (ArrayUtils.isNotEmpty(prop.generated())) {
                property.extra.put(BeanProps.GENERATED, MergedAnnotations.from(prop.generated()).get(Generation.class).asMap());
                if (StringUtils.isBlank((String) property.extra.get(BeanProps.COMPONENT))) {
                    property.extra.put(BeanProps.COMPONENT, "input-hidden");
                }
            } else {
                property.extra.remove(BeanProps.GENERATED);
                if ("input-hidden".equals(property.extra.get(BeanProps.COMPONENT))) {
                    property.extra.remove(BeanProps.COMPONENT);
                }
            }
            // 设置标题
            property.title = prop.title();
        } else {
            BeanProperty property = new BeanProperty();
            property.setType(BeanPropertyType.STRING);
            property.setName(key);
            property.setTitle(prop.title());
            if (!prop.readonly()) {
                result.add(property);
            }
        }
    }

    /**
     * 替换文案的生效
     *
     * @param property    属性
     * @param annotations 注解
     */
    private static void applyReplacement(BeanProperty property, MergedAnnotations annotations) {
        SubBean subBean = annotations.get(SubBean.class).synthesize();
        if (CollectionUtils.isNotEmpty(property.getChildren()) && StringUtils.isNotBlank(subBean.search())
                && StringUtils.isNotBlank(subBean.replacement())) {
            property.getChildren().forEach(child -> child.setTitle(child.getTitle()
                    .replace(subBean.search(), subBean.replacement())));
        }
    }

    /**
     * 设置当前对象的键值属性
     *
     * @param key   键
     * @param value 值
     * @return 结果
     */
    public BeanProperty prop(String key, Object value) {
        if (null == props) {
            props = new HashMap<>();
        }
        props.put(key, value);
        return this;
    }

    /**
     * 设置当前对象的额外属性
     *
     * @param key   键
     * @param value 值
     * @return 结果
     */
    public BeanProperty option(String key, Object value) {
        if (null == options) {
            options = new HashMap<>();
        }
        options.put(key, value);
        return this;
    }

    private boolean isAttachment(Class<?> clazz) {
        return "Attachment".equals(clazz.getSimpleName());
    }
}
