package com.flyfish.framework.configuration.jwt;

import com.flyfish.framework.service.MongoUserDetailsService;
import lombok.extern.slf4j.Slf4j;
import org.springframework.http.server.reactive.ServerHttpRequest;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.web.server.context.ServerSecurityContextRepository;
import org.springframework.web.server.ServerWebExchange;
import reactor.core.publisher.Mono;

import javax.annotation.Resource;
import java.net.URI;

@Slf4j
public class JwtSecurityContextRepository implements ServerSecurityContextRepository {

    @Resource
    private MongoUserDetailsService userDetailsService;
    @Resource
    private TokenProvider tokenProvider;

    @Override
    public Mono<SecurityContext> load(ServerWebExchange serverWebExchange) {
        ServerHttpRequest request = serverWebExchange.getRequest();
        String jwt = tokenProvider.retrieveToken(serverWebExchange).orElse(null);
        URI requestURI = request.getURI();
        // 存在jwt时，校验jwt。redis也需要存在
        return tokenProvider.validateToken(jwt).flatMap(authorized -> {
            if (authorized) {
                // token即将过期，续租
                Authentication authentication = tokenProvider.getAuthentication(jwt);
                log.debug("set Authentication to security context for '{}', uri: {}", authentication.getName(), requestURI);
                return userDetailsService.findByUsername(authentication.getName())
                        .flatMap(userDetails -> userDetailsService.loadContext(userDetails));
            } else {
                log.debug("no valid JWT token found, uri: {}", requestURI);
                return Mono.empty();
            }
        });
    }

    /**
     * 只有登录会调用这个
     *
     * @param webExchange     交换信息
     * @param securityContext 上下文
     * @return 结果
     */
    @Override
    public Mono<Void> save(ServerWebExchange webExchange, SecurityContext securityContext) {
        // 添加jwtToken
        tokenProvider.addToken(webExchange, securityContext.getAuthentication());
        return Mono.empty();
    }
}
